package uulm.teamname.marvelous.server;

import org.java_websocket.WebSocket;
import org.junit.jupiter.api.*;
import org.mockito.MockedStatic;
import org.mockito.Mockito;
import org.mockito.invocation.Invocation;
import org.tinylog.configuration.Configuration;
import uulm.teamname.marvelous.gamelibrary.json.JSON;
import uulm.teamname.marvelous.gamelibrary.messages.BasicMessage;
import uulm.teamname.marvelous.gamelibrary.messages.ParticipantType;
import uulm.teamname.marvelous.gamelibrary.messages.RoleEnum;
import uulm.teamname.marvelous.gamelibrary.messages.client.CharacterSelectionMessage;
import uulm.teamname.marvelous.gamelibrary.messages.client.HelloServerMessage;
import uulm.teamname.marvelous.gamelibrary.messages.client.PlayerReadyMessage;
import uulm.teamname.marvelous.gamelibrary.messages.client.ReconnectMessage;
import uulm.teamname.marvelous.gamelibrary.messages.server.*;
import uulm.teamname.marvelous.server.lobbymanager.LobbyConnection;
import uulm.teamname.marvelous.server.lobbymanager.LobbyManager;
import uulm.teamname.marvelous.server.lobbymanager.LobbyRunner;
import uulm.teamname.marvelous.server.netconnector.UserManager;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import static org.mockito.Mockito.*;

class MarvelousServerTest extends BaseGameLogicTest {
    private static MockedStatic<Server> serverMock;
    private static JSON json;

    @BeforeAll
    static void start() throws NoSuchMethodException, IllegalAccessException, InvocationTargetException, InstantiationException, NoSuchFieldException {
        Constructor<UserManager> um = UserManager.class.getDeclaredConstructor();
        um.setAccessible(true);
        um.newInstance();
        Constructor<LobbyManager> lm = LobbyManager.class.getDeclaredConstructor();
        lm.setAccessible(true);
        lm.newInstance();

        serverMock = Mockito.mockStatic(Server.class);

        generate();

        setPrivateFinalBoolean(LobbyConnection.class, "synchronous", true);
        setPrivateFinalBoolean(LobbyRunner.class, "synchronous", true);

        serverMock.when(Server::getMaxLobbies).thenReturn(10);
        serverMock.when(Server::getPartyConfig).thenReturn(partyConfig);
        serverMock.when(Server::getScenarioConfig).thenReturn(scenarioConfig);
        serverMock.when(Server::getCharacterConfig).thenReturn(characterConfig);

        json = new JSON(characterConfig);

        if (!Configuration.isFrozen()) {
            Configuration.set("writer1", "console");
            Configuration.set("writer1.level", "trace");
            Configuration.set("writer1.format", "[{thread}] {level}: <TEST> {message}");
        }
   }


    @Test
    void main() {
        UserManager m = UserManager.getInstance();

        WebSocket p1 = mock(WebSocket.class);
        WebSocket p2 = mock(WebSocket.class);
        WebSocket s1 = mock(WebSocket.class);
        WebSocket s2 = mock(WebSocket.class);

        when(p1.getResourceDescriptor()).thenReturn("/");
        when(p2.getResourceDescriptor()).thenReturn("/");
        when(s1.getResourceDescriptor()).thenReturn("/");
        when(s2.getResourceDescriptor()).thenReturn("/");

        m.connectUser(p1);
        ensureHandshake(m, p1, "Player 1", "1234", false);

        m.connectUser(p2);
        ensureHandshake(m, p2, "Player 2", "4321", false);

        m.connectUser(s1);
        ensureHandshake(m, s1, "Spectator 1", "3333", false);

        m.connectUser(s2);
        ensureHandshake(m, s2, "Spectator 2", "4444", false);

        ensurePlayerReady(m, p1, true, RoleEnum.PLAYER);
        ensurePlayerReady(m, p2, true, RoleEnum.PLAYER);
        ensureSpectatorReady(m, s1, true, RoleEnum.SPECTATOR);

        ensureCharacterSelection(m, p1, true);
        ensureCharacterSelection(m, p2, false);

        GameStructureMessage game = new GameStructureMessage();
        game.playerOneName = "Player 1";
        game.playerTwoName = "Player 2";
        game.matchconfig = partyConfig;
        game.scenarioconfig = scenarioConfig;

        game.assignment = ParticipantType.PlayerOne;
        ensureReceived(p1, game);

        game.assignment = ParticipantType.PlayerTwo;
        ensureReceived(p2, game);

        game.assignment = ParticipantType.Spectator;
        ensureReceived(s1, game);

        EventMessage events = new EventMessage();
        ensureReceived(p1, events);
        ensureReceived(p2, events);
        ensureReceived(s1, events);

        clearInvocations(p1, p2, s1);

        m.disconnectUser(p1, true);

        m.connectUser(p1);
        ensureHandshake(m, p1, "Player 1", "1234", true);

        ReconnectMessage message = new ReconnectMessage();
        message.reconnect = true;
        sendMessage(m, p1, message);

        ensureReceived(p1, new GeneralAssignmentMessage());
        ensureReceived(p1, new GameStructureMessage());

        System.out.println("Test Completed");
    }

    private void ensureHandshake(UserManager m, WebSocket c, String name, String deviceID, boolean runningGame) {
        HelloServerMessage message = new HelloServerMessage();
        message.name = name;
        message.deviceID = deviceID;

        HelloClientMessage response = new HelloClientMessage();
        response.runningGame = runningGame;

        ensureResponse(m, c, message, response);
    }

    private void ensurePlayerReady(UserManager m, WebSocket c, boolean startGame, RoleEnum role) {
        PlayerReadyMessage message = new PlayerReadyMessage();
        message.startGame = startGame;
        message.role = role;

        GameAssignmentMessage response = new GameAssignmentMessage();
        //properties are left null because we can't test their content (they are random)

        ensureResponse(m, c, message, response);
    }

    private void ensureSpectatorReady(UserManager m, WebSocket c, boolean startGame, RoleEnum role) {
        PlayerReadyMessage message = new PlayerReadyMessage();
        message.startGame = startGame;
        message.role = role;

        GeneralAssignmentMessage response = new GeneralAssignmentMessage();
        //properties are left null because we can't test their content (they are random)

        ensureResponse(m, c, message, response);
    }

    private void ensureCharacterSelection(UserManager m, WebSocket c, boolean confirmSelection) {
        CharacterSelectionMessage message = new CharacterSelectionMessage();
        message.characters = new Boolean[12];
        for(int i = 0; i < 6; i++) {
            message.characters[i] = true;
        }
        for(int i = 6; i < 12; i++) {
            message.characters[i] = false;
        }

        if(confirmSelection) {
            ConfirmSelectionMessage response = new ConfirmSelectionMessage();
            response.selectionComplete = false;
            ensureResponse(m, c, message, response);
        } else {
            sendMessage(m, c, message);
        }
    }


    /** Sends a message from the socket. */
    private void sendMessage(UserManager m, WebSocket c, BasicMessage message) {
        Optional<String> in = json.stringify(message);

        if(in.isPresent()) {
            m.messageReceived(c, in.get());
        }else {
            throw new IllegalArgumentException("[TEST] Message in test call could not be serialized!");
        }
    }

    /** Ensures that the given socket received the given response in response to the given message. */
    private void ensureResponse(UserManager m, WebSocket c, BasicMessage message, BasicMessage response) {
        Optional<String> in = json.stringify(message);

        if(in.isPresent()) {
            m.messageReceived(c, in.get());
            verify(c).send((String)argThat(
                    (out)->verifyResponse((String)out, response)
            ));
            clearInvocations(c);
        }else {
            throw new IllegalArgumentException("[TEST] Message in test call could not be serialized!");
        }
    }

    /** Ensures that the given socket received the given response. */
    private <T extends BasicMessage> void ensureReceived(WebSocket c, T response) {
        boolean found = false;

        for(Invocation i: mockingDetails(c).getInvocations()) {
            Optional<BasicMessage> received = json.parse(i.getArgument(0, String.class));

            if(received.isPresent()) {
                BasicMessage parsed = received.get();

                if(parsed.getClass() != response.getClass()) {
                    continue; //message is not of the right type
                }

                T message = (T) parsed;

                for(Field field: response.getClass().getDeclaredFields()) {
                    try {
                        Object value = field.get(response);
                        Object actualValue;
                        try {
                            actualValue = field.get(message);
                        }catch(IllegalArgumentException ignored) {
                            continue;
                        }
                        if(value != null && !value.equals(actualValue)) {
                            throw new IllegalArgumentException("[TEST] Field " + field.getName() + " expected to be '" + value + "' but was '" + field.get(message) + "'");
                        }
                    }catch(IllegalAccessException ignored) { }
                }

                found = true;
                break;
            }else {
                throw new IllegalArgumentException("[TEST] Response in test call could not be deserialized!");
            }
        }

        if(!found) {
            throw new IllegalArgumentException("[TEST] Expected response " + response + " was not received!");
        }
    }


    /**
     * Verifies a response message and requires all non-null fields of the expected message
     * to equal the respective values in the actual received message.
     */
    private <T extends BasicMessage> boolean verifyResponse(String actual, T expected) {
        Optional<BasicMessage> in = json.parse(actual);

        if(in.isPresent()) {
            T message = (T) in.get();

            for(Field field: expected.getClass().getDeclaredFields()) {
                try {
                    Object value = field.get(expected);
                    Object actualValue;
                    try {
                        actualValue = field.get(message);
                    }catch(IllegalArgumentException e) {
                        throw new IllegalArgumentException("[TEST] Response message expected to be '" + expected.getClass().getName() + "' but was '" + message.getClass().getName() + "'");
                    }
                    if(value != null && !value.equals(actualValue)) {
                        throw new IllegalArgumentException("[TEST] Field " + field.getName() + " expected to be '" + value + "' but was '" + field.get(message) + "'");
                    }
                }catch(IllegalAccessException ignore) { }
            }

            return true;
        }else {
            throw new IllegalArgumentException("[TEST] Response in test call could not be deserialized!");
        }
    }

    @AfterAll
    static void stop() {
        serverMock.close();
    }
}
